package dk.apaq.framework.repository.jpa;

import java.lang.reflect.Field;
import java.util.List;
import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;
import javax.persistence.Id;
import javax.persistence.Query;
import dk.apaq.framework.criteria.Criteria;
import dk.apaq.framework.criteria.jpa.CriteriaForJpa;
import dk.apaq.framework.repository.BaseRepository;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 */
public class EntityManagerRepository<BEANTYPE, IDTYPE> extends BaseRepository<BEANTYPE, IDTYPE> {

    private static final Logger LOG = LoggerFactory.getLogger(EntityManagerRepository.class);
    private final EntityManager entityManager;
    private final Class<BEANTYPE> clazz;
    private boolean transactionsHandled = true;

    public EntityManagerRepository(EntityManager entityManager, Class<BEANTYPE> clazz) {
        this.entityManager = entityManager;
        this.clazz = clazz;
        
        if(entityManager==null) { throw new NullPointerException("EntityManager was null"); }
        if(clazz==null) { throw new NullPointerException("Class was null"); }
        
    }

    public IDTYPE getIdFromEntity(BEANTYPE entity) {
        return resolveId(entity);
    }
    
    public BEANTYPE findOne(IDTYPE id) {
        fireOnBeforeRead(id);
        BEANTYPE entity = doRead(id);
        fireOnRead(id, entity);
        return entity;
    }

    public List<IDTYPE> findAllIds(Criteria criteria) {
        fireOnBeforeList(criteria);
        Query q = CriteriaForJpa.translate(entityManager, new String[]{getIdProperty()}, clazz, criteria);
        return q.getResultList();
    }
    
    public List<BEANTYPE> findAll(Criteria criteria) {
        fireOnBeforeList(criteria);
        Query q = CriteriaForJpa.translate(entityManager, clazz, criteria);
        return q.getResultList();
    }

    public BEANTYPE save(final BEANTYPE entity) {
        IDTYPE id = resolveId(entity);
        if(id == null) {
            fireOnBeforeCreate(entity);
            final Object[] entityA = new Object[1];
            runInTransaction(new Runnable() {

                public void run() {
                    entityA[0] = entityManager.merge(entity);
                    entityManager.flush();
                }
            });

            id = resolveId((BEANTYPE) entityA[0]);
            fireOnCreate(id, entity);
            return (BEANTYPE) entityA[0];
        } else {
            fireOnBeforeUpdate(id, entity);
            runInTransaction(new Runnable() {

                public void run() {
                    entityManager.merge(entity);
                    entityManager.flush();
                }
            });
            BEANTYPE mergedEntity = entityManager.merge(entity);
            fireOnUpdate(id, mergedEntity);
            return mergedEntity;
        }
    }

    public void delete(BEANTYPE entity) {
        IDTYPE id = resolveId(entity);
        deleteById(id);
    }

    public void deleteById(IDTYPE id) {
        fireOnBeforeDelete(id);
        final BEANTYPE entity = doRead(id);
        runInTransaction(new Runnable() {

            public void run() {
                entityManager.remove(entity);
                entityManager.flush();
            }
        });
        fireOnDelete(id);
    }

    private BEANTYPE doRead(IDTYPE id) {
        return entityManager.find(clazz, id);
    }

    private String getIdProperty() {
        Field field = getIdField();
        if (field != null) {
            return field.getName();
        }
        return null;
    }

    private IDTYPE resolveId(final BEANTYPE entity) {
        Field field = getIdField();
        if(field==null) {
            LOG.warn("Unable to find Id annotated field in entity.");
            return null;
        }
        
        field.setAccessible(true);
        try {
            return (IDTYPE) field.get(entity);
        } catch (Exception ex) {
            LOG.error("Unable to retrieve value from Id field.", ex);
            return null;
        }
    }

    private Field getIdField() {
        Class currentClass = this.clazz;

        while (currentClass != null) {
            for (Field field : currentClass.getDeclaredFields()) {
                if (field.getAnnotation(Id.class) != null) {
                    return field;
                }
            }
            currentClass = currentClass.getSuperclass();
        }
        return null;
    }

    /**
     * Specifies whether the entity provider should handle transactions itself
     * or whether they should be handled outside (e.g. if declarative
     * transactions are used).
     *
     * @param transactionsHandled
     *            true to handle the transactions internally, false to rely on
     *            external transaction handling.
     */
    public void setTransactionsHandledByProvider(boolean transactionsHandled) {
        this.transactionsHandled = transactionsHandled;
    }

    /**
     * Returns whether the entity provider is handling transactions internally
     * (the default) or relies on external transaction handling.
     *
     * @return true if transactions are handled internally, false if not.
     */
    public boolean isTransactionsHandledByProvider() {
        return transactionsHandled;
    }

    /**
     * If {@link #isTransactionsHandledByProvider() } is true,
     * <code>operation</code> will be executed inside a transaction that is
     * commited after the operation is completed. Otherwise,
     * <code>operation</code> will just be executed.
     *
     * @param operation the operation to run (must not be null).
     */
    protected void runInTransaction(Runnable operation) {
        assert operation != null : "operation must not be null";
        if (isTransactionsHandledByProvider()) {
            EntityTransaction et = entityManager.getTransaction();
            if (et.isActive()) {
                // The transaction has been started outside of this method
                // and should also be committed/rolled back outside of
                // this method
                operation.run();
            } else {
                try {
                    et.begin();
                    operation.run();
                    et.commit();
                } finally {
                    if (et.isActive()) {
                        et.rollback();
                    }
                }
            }
        } else {
            operation.run();
        }
    }
}
